//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_ID3V20_NUM_HPP_
#define _LFC1_ID3V20_NUM_HPP_

#include <type_traits>
#include <limits>
#include <cmath>
#include <algorithm>

namespace lfc1 { namespace id3v20 {

///
///	@ingroup LFC1_ID3V20
///
///	@brief This template class represents an ID3 v2.x unsigned integer.
///
///	This class supports the following operators:
///	@code
///	binary arithmetic operators: +=, -=, *=, /=
///	unary arithmetic operators: ++, --
///	@endcode
///
///	@tparam	Bytes		The number of bytes in the integer.
///	@tparam	Uint_type	The equivalent unsigned integer type.
///
template <int Bytes, typename Uint_type = unsigned>
// requires
class num_c final
{
private:
	static_assert(Bytes >= 1 && Bytes <= 4, "Number of bytes must be in the range 1 to 4");
	static_assert
	(
		std::is_integral<Uint_type>::value && std::is_unsigned<Uint_type>::value,
		"Uint_type must be an unsigned integral type"
	);

public:
	// Constructors/destructor
	///
	///	@brief Creates a lfc1::id3v20::num_c object.
	///
	num_c() noexcept = default;
	///
	///	@brief Creates a lfc1::id3v20::num_c object.
	///
	///	@param[in]	integer		The value to assign to the lfc1::id3v20::num_c object.
	///
	num_c(Uint_type integer) noexcept;

	// Assignment
	///
	///	@brief Assigns the given value to a lfc1::id3v20::num_c object.
	///
	///	@param[in]	integer		The value to assign to the lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object assigned to.
	///
	num_c& operator=(Uint_type integer) noexcept;

	// Conversions
	///
	///	@brief Converts a lfc1::id3v20::num_c object to its corresponding integer type.
	///
	///	@return	The value of the integer type.
	///
	operator Uint_type () const noexcept;

	// Instance setters/getters
	///
	///	@brief Obtains the size of a lfc1::id3v20::num_c object when stored in a file.
	///
	///	@return The size of a lfc1::id3v20::num_c object when stored in a file.
	///
	int get_size() const noexcept;

	// Instance special operations
	///
	///	@brief Swaps a lfc1::id3v20::num_c object with another lfc1::id3v20::num_c object.
	///
	///	@param[in,out]	rhs		The lfc1::id3v20::num_c object to swap with.
	///
	void swap(num_c& rhs) noexcept;
	///
	///	@brief Overloads the pre-increment operator for lfc1::id3v20::num_c objects.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator++() noexcept;
	///
	///	@brief Overloads the pre-decrement operator for lfc1::id3v20::num_c objects.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator--() noexcept;
	///
	///	@brief Overloads the add equal operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator+=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the subtract equal operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator-=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the multiply equal operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator*=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the divide equal operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator/=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the logical and operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator&=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the logical or operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator|=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the logical exclusive or operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator^=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the left shift operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator<<=(const num_c& rhs) noexcept;
	///
	///	@brief Overloads the right shift operator for lfc1::id3v20::num_c objects.
	///
	///	@param[in]	rhs		The other lfc1::id3v20::num_c object.
	///
	///	@return	The lfc1::id3v20::num_c object.
	///
	num_c& operator>>=(const num_c& rhs) noexcept;

private:
	// Instance data
	unsigned char data_[Bytes];
};

///
///	@ingroup LFC1_ID3V20
///
///	@brief Overloads std::swap for lfc1::id3v20::num_c objects.
///
///	@tparam	Bytes		The number of bytes in the integer.
///	@tparam	Uint_type	The equivalent unsigned integer type.
///
///	@param[in,out]	lhs		The first object to swap.
///	@param[in,out]	rhs		The second object to swap.
///
template <int Bytes, typename Uint_type = unsigned>
void swap(num_c<Bytes, Uint_type>& lhs, num_c<Bytes, Uint_type>& rhs) noexcept;
///
///	@ingroup LFC1_ID3V20
///
///	@brief Overloads the post-increment operator for lfc1::id3v20::num_c objects.
///
///	@tparam	Bytes		The number of bytes in the integer.
///	@tparam	Uint_type	The equivalent unsigned integer type.
///
///	@param[in,out]	num		The lfc1::id3v20::num_c object to increment.
///
///	@return	The value of the lfc1::id3v20::num_c object before the increment.
///
template <int Bytes, typename Uint_type = unsigned>
num_c<Bytes, Uint_type> operator++(num_c<Bytes, Uint_type>& num, int) noexcept;
///
///	@ingroup LFC1_ID3V20
///
///	@brief Overloads the post-decrement operator for lfc1::id3v20::num_c objects.
///
///	@tparam	Bytes		The number of bytes in the integer.
///	@tparam	Uint_type	The equivalent unsigned integer type.
///
///	@param[in,out]	num		The lfc1::id3v20::num_c object to decrement.
///
///	@return	The value of the lfc1::id3v20::num_c object before the decrement.
///
template <int Bytes, typename Uint_type = unsigned>
num_c<Bytes, Uint_type> operator--(num_c<Bytes, Uint_type>& num, int) noexcept;

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>::num_c(Uint_type integer) noexcept
{
	// Store the value.
	auto data = reinterpret_cast<unsigned char*>(&data_);
	for (int i = sizeof(data_) - 1; i >= 0; --i)
	{
		data[i] = static_cast<unsigned char>(integer & 0xFF);
		integer >>= 8;
	}
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator=(Uint_type integer) noexcept
{
	// Store the value.
	auto data = reinterpret_cast<unsigned char*>(&data_);
	for (int i = sizeof(data_) - 1; i >= 0; --i)
	{
		data[i] = static_cast<unsigned char>(integer & 0xFF);
		integer >>= 8;
	}
	return *this;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>::operator Uint_type () const noexcept
{
	Uint_type integer = 0;
	auto data = reinterpret_cast<const unsigned char*>(&data_);
	for (auto i = 0; i < static_cast<decltype(i)>(sizeof(data_)); ++i)
	{
		integer = (integer << 8) + data[i];
	}
	return integer;
}

template <int Bytes, typename Uint_type>
int num_c<Bytes, Uint_type>::get_size() const noexcept
{
	return Bytes;
}

template <int Bytes, typename Uint_type>
void num_c<Bytes, Uint_type>::swap(num_c& rhs) noexcept
{
	using std::swap;
	swap(data_, rhs.data_);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator++() noexcept
{
	Uint_type integer = *this;
	return (*this = integer + 1);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator--() noexcept
{
	Uint_type integer = *this;
	return (*this = integer - 1);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator+=(const num_c& rhs) noexcept
{
	Uint_type integer = *this + rhs;
	return (*this = integer);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator-=(const num_c& rhs) noexcept
{
	Uint_type integer = *this - rhs;
	return (*this = integer);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator*=(const num_c& rhs) noexcept
{
	Uint_type integer = *this * rhs;
	return (*this = integer);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator/=(const num_c& rhs) noexcept
{
	Uint_type integer = *this / rhs;
	return (*this = integer);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator&=(const num_c& rhs) noexcept
{
	Uint_type integer = *this & rhs;
	*this = integer;
	return *this;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator|=(const num_c& rhs) noexcept
{
	Uint_type integer = *this | rhs;
	*this = integer;
	return *this;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator^=(const num_c& rhs) noexcept
{
	Uint_type integer = *this ^ rhs;
	*this = integer;
	return *this;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator<<=(const num_c& rhs) noexcept
{
	Uint_type integer = *this << rhs;
	*this = integer;
	return *this;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type>& num_c<Bytes, Uint_type>::operator>>=(const num_c& rhs) noexcept
{
	Uint_type integer = *this >> rhs;
	*this = integer;
	return *this;
}

template <int Bytes, typename Uint_type>
void swap(num_c<Bytes, Uint_type>& lhs, num_c<Bytes, Uint_type>& rhs) noexcept
{
	lhs.swap(rhs);
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type> operator++(num_c<Bytes, Uint_type>& num, int) noexcept
{
	Uint_type integer = num;
	++num;
	return integer;
}

template <int Bytes, typename Uint_type>
num_c<Bytes, Uint_type> operator--(num_c<Bytes, Uint_type>& num, int) noexcept
{
	Uint_type integer = num;
	--num;
	return integer;
}

} /* namespace id3v20 */ } /* namespace lfc1 */

namespace std {

///
///	@ingroup LFC1_ID3V20
///
///	@brief Specializes std::numeric_limits for lfc1::id3v20::num_c objects.
///
///	@tparam	Bytes		The number of bytes in the integer.
///	@tparam	Uint_type	The equivalent unsigned integer type.
///
template <int Bytes, typename Uint_type>
class numeric_limits<lfc1::id3v20::num_c<Bytes, Uint_type>>
{
public:
	///
	///	@brief Have information for this type.
	///
	static constexpr bool is_specialized = true;
	///
	///	@brief Obtains the minimum finite value for a lfc1::id3v20::num_c.
	///
	///	@return The minimum finite value for a lfc1::id3v20::num_c.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> min() noexcept
	{
		return 0;
	}
	///
	///	@brief Obtains the maximum finite value for a lfc1::id3v20::num_c.
	///
	///	@return The maximum finite value for a lfc1::id3v20::num_c.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> max() noexcept
	{
		return static_cast<unsigned>(std::pow(2.0, 8.0 * Bytes) - 1);
	}
	///
	///	@brief Obtains the lowest finite value for a lfc1::id3v20::num_c.
	///
	///	@return The lowest finite value for a lfc1::id3v20::num_c.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> lowest() noexcept
	{
		return 0;
	}
	///
	///	@brief Number of radix digits that can be represented without change in a lfc1::id3v20::num_c.
	///
	static constexpr int digits = 8 * Bytes;
	///
	///	@brief Number of base 10 digits that can be represented without change in a lfc1::id3v20::num_c.
	///
#if defined(__FEDORA__)
	static constexpr int digits10 = static_cast<int>(std::log10(std::pow(2.0, 8.0 * Bytes)));
#else
	static constexpr int digits10 = ((Bytes == 1) ? 2 : ((Bytes == 2) ? 4 : ((Bytes == 3) ? 7 : 9)));
#endif
	///
	///	@brief Number of base 10 digits required to ensure that values that differ are always differentiated
	///		   in a lfc1::id3v20::num_c.
	///
	static constexpr int max_digits10 = 0;
	///
	///	@brief Indicates whether a lfc1::id3v20::num_c is signed or not.
	///
	static constexpr bool is_signed = false;
	///
	///	@brief Indicates whether a lfc1::id3v20::num_c is an integer or not.
	///
	static constexpr bool is_integer = true;
	///
	///	@brief Indicates whether a lfc1::id3v20::num_c uses an exact representation or not.
	///
	static constexpr bool is_exact = true;
	///
	///	@brief Specifies the base of the representation of a lfc1::id3v20::num_c.
	///
	static constexpr int radix = 2;
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> epsilon() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> round_error() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	static constexpr int min_exponent = 0;
	///
	///	@brief Not applicable.
	///
	static constexpr int min_exponent10 = 0;
	///
	///	@brief Not applicable.
	///
	static constexpr int max_exponent = 0;
	///
	///	@brief Not applicable.
	///
	static constexpr int max_exponent10 = 0;
	///
	///	@brief Not applicable.
	///
	static constexpr bool has_infinity = false;
	///
	///	@brief Not applicable.
	///
	static constexpr bool has_quiet_NaN = false;
	///
	///	@brief Not applicable.
	///
	static constexpr bool has_signaling_NaN = false;
	///
	///	@brief Not applicable.
	///
	static constexpr float_denorm_style has_denorm = denorm_absent;
	///
	///	@brief Not applicable.
	///
	static constexpr bool has_denorm_loss = false;
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> infinity() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> quiet_NaN() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> signaling_NaN() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	///	@return Do not care.
	///
	static constexpr lfc1::id3v20::num_c<Bytes, Uint_type> denorm_min() noexcept { return 0; }
	///
	///	@brief Not applicable.
	///
	static constexpr bool is_iec559 = false;
	///
	///	@brief lfc1::id3v20::num_c has a range of valid values.
	///
	static constexpr bool is_bounded = true;
	///
	///	@brief lfc1::id3v20::num_c values wrap around on underflow/overflow.
	///
	static constexpr bool is_modulo = true;
	///
	///	@brief Not applicable.
	///
	static constexpr bool traps = false;
	///
	///	@brief Not applicable.
	///
	static constexpr bool tinyness_before = false;
	///
	///	@brief Not applicable.
	///
	static constexpr float_round_style round_style = round_toward_zero;
};

} /* namespace std */

#endif // _LFC1_ID3V20_NUM_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
